-- action bars
-- bars that contain ... actions

local AddonName, Addon = ...

local MAX_BUTTONS = #Addon.ActionButtons
local ACTION_BUTTON_SHOW_GRID_REASON_ADDON = 1024
local ACTION_BUTTON_SHOW_GRID_REASON_KEYBOUND = 2048

local ActionBar = Addon:CreateClass('Frame', Addon.ButtonBar)
Addon.ActionBar = ActionBar

-- Metatable magic.  Basically this says, 'create a new table for this index'
-- I do this so that I only create page tables for classes the user is actually
-- playing
ActionBar.defaultOffsets = {
    __index = function(t, i)
        t[i] = {}
        return t[i]
    end
}

-- Metatable magic.  Basically this says, 'create a new table for this index,
-- with these defaults. I do this so that I only create page tables for classes
-- the user is actually playing
ActionBar.mainbarOffsets = {
    __index = function(t, i)
        local pages = {
            page2 = 1,
            page3 = 2,
            page4 = 3,
            page5 = 4,
            page6 = 5
        }

        if i == 'DRUID' then
            pages.cat = 6
            pages.bear = 8
            pages.moonkin = 9
            pages.tree = 7
        elseif i == 'ROGUE' then
            pages.stealth = 6
            pages.shadowdance = 6
        elseif i == 'WARRIOR' and Addon:IsBuild('classic') then
            pages.battle = 6
            pages.defensive = 7
            pages.berserker = 8
        elseif i == 'PRIEST' and Addon:IsBuild('classic') then
            pages.shadowform = 6
        end

        t[i] = pages
        return pages
    end
}

ActionBar.class = select(2, UnitClass('player'))

local active = {}

function ActionBar:New(id)
    local bar = ActionBar.proto.New(self, id)

    bar.sets.pages = setmetatable(bar.sets.pages, bar.id == 1 and self.mainbarOffsets or self.defaultOffsets)
    bar.pages = bar.sets.pages[bar.class]

    bar:LoadStateController()
    bar:UpdateStateDriver()
    bar:SetRightClickUnit(Addon:GetRightClickUnit())
    bar:UpdateGrid()
    bar:UpdateTransparent(true)

    active[id] = bar

    return bar
end

--TODO: change the position code to be based more on the number of action bars
function ActionBar:GetDefaults()
    return {
        point = 'BOTTOM',
        x = 0,
        y = 40 * (self.id - 1),
        pages = {},
        spacing = 4,
        padW = 2,
        padH = 2,
        numButtons = self:MaxLength()
    }
end

function ActionBar:GetDisplayName()
    local L = LibStub('AceLocale-3.0'):GetLocale(AddonName)

    return L.ActionBarDisplayName:format(self.id)
end

function ActionBar:Free()
    active[self.id] = nil

    ActionBar.proto.Free(self)
end

--returns the maximum possible size for a given bar
function ActionBar:MaxLength()
    return floor(MAX_BUTTONS / Addon:NumBars())
end

function ActionBar:AcquireButton(index)
    local actionID = index + (self.id - 1) * self:MaxLength()
    local button = Addon.ActionButtons[actionID]

    button:SetAttribute('index', index)
    button:SetAttribute('statehidden', nil)

    return button
end

function ActionBar:ReleaseButton(button)
    button:SetAttribute('statehidden', true)
    button:SetParent(Addon.ShadowUIParent)
    button:Hide()
end

function ActionBar:OnAttachButton(button)
    button:SetActionOffsetInsecure(self:GetAttribute("state-offset"))

    button:SetFlyoutDirection(self:GetFlyoutDirection())
    button:SetShowCountText(Addon:ShowCounts())
    button:SetShowMacroText(Addon:ShowMacroText())
    button:SetShowEquippedItemBorders(Addon:ShowEquippedItemBorders())
    button:SetShowCooldowns(self:GetAlpha() > 0)
    button:UpdateHotkeys()

    Addon:GetModule('ButtonThemer'):Register(button, self:GetDisplayName())
    Addon:GetModule('Tooltips'):Register(button)
end

function ActionBar:OnDetachButton(button)
    Addon:GetModule('ButtonThemer'):Unregister(button, self:GetDisplayName())
    Addon:GetModule('Tooltips'):Unregister(button)
end

-- paging
function ActionBar:SetOffset(stateId, page)
    self.pages[stateId] = page
    self:UpdateStateDriver()
end

function ActionBar:GetOffset(stateId)
    return self.pages[stateId]
end

function ActionBar:UpdateStateDriver()
    local conditions

    for _, state in Addon.BarStates:getAll() do
        local offset = self:GetOffset(state.id)

        if offset then
            local condition

            if type(state.value) == 'function' then
                condition = state.value()
            else
                condition = state.value
            end

            if condition then
                local page = Wrap(self.id + offset, Addon:NumBars())

                if conditions then
                    conditions = strjoin(';', conditions, (condition .. page))
                else
                    conditions = (condition .. page)
                end
            end
        end
    end

    if conditions then
        RegisterStateDriver(self, 'page', strjoin(';', conditions, self.id))
    else
        UnregisterStateDriver(self, 'page')
        self:SetAttribute('state-page', self.id)
    end
end

function ActionBar:LoadStateController()
    self:SetAttribute('barLength', self:MaxLength())
    self:SetAttribute('overrideBarLength', NUM_ACTIONBAR_BUTTONS)

    self:SetAttribute('_onstate-overridebar', [[ self:RunAttribute('UpdateOffset') ]])
    self:SetAttribute('_onstate-overridepage', [[ self:RunAttribute('UpdateOffset') ]])
    self:SetAttribute('_onstate-page', [[ self:RunAttribute('UpdateOffset') ]])

    self:SetAttribute(
        'UpdateOffset',
        [[
            local offset = 0

            local overridePage = self:GetAttribute('state-overridepage') or 0
            if overridePage > 10 and self:GetAttribute('state-overridebar') then
                offset = (overridePage - 1) * self:GetAttribute('overrideBarLength')
            else
                local page = self:GetAttribute('state-page') or 1
                offset = (page - 1) * self:GetAttribute('barLength')
            end

            control:ChildUpdate('offset', offset)
        ]]
    )

    self:UpdateOverrideBar()
end

function ActionBar:UpdateOverrideBar()
    local isOverrideBar = self:IsOverrideBar()

    self:SetAttribute('state-overridebar', isOverrideBar)
end

--returns true if the possess bar, false otherwise
function ActionBar:IsOverrideBar()
    return Addon:GetOverrideBar() == self
end

--Empty button display
function ActionBar:ShowGrid(reason)
    if InCombatLockdown() then
        return
    end

    self:ForButtons('ShowGridInsecure', reason)
end

function ActionBar:HideGrid(reason)
    if InCombatLockdown() then
        return
    end

    self:ForButtons('HideGridInsecure', reason)
end

function ActionBar:UpdateGrid()
    if Addon:ShowGrid() then
        self:ShowGrid(ACTION_BUTTON_SHOW_GRID_REASON_ADDON)
    else
        self:HideGrid(ACTION_BUTTON_SHOW_GRID_REASON_ADDON)
    end
end

---keybound support
function ActionBar:KEYBOUND_ENABLED()
    self:ShowGrid(ACTION_BUTTON_SHOW_GRID_REASON_KEYBOUND)
end

function ActionBar:KEYBOUND_DISABLED()
    self:HideGrid(ACTION_BUTTON_SHOW_GRID_REASON_KEYBOUND)
end

-- right click targeting support
function ActionBar:SetRightClickUnit(unit)
    unit = unit or 'none'

    if unit == 'none' then
        self:SetAttribute('*unit2', nil)
    else
        self:SetAttribute('*unit2', unit)
    end
end

function ActionBar:GetRightClickUnit()
    return self:SetAttribute('*unit2') or 'none'
end

function ActionBar:OnSetAlpha(alpha)
    self:UpdateTransparent()
end

function ActionBar:UpdateTransparent(force)
    local isTransparent = self:GetAlpha() == 0

    if self.__transparent ~= isTransparent or force then
        self.__transparent = isTransparent

        self:ForButtons('SetShowCooldowns', not isTransparent)
    end
end

-- flyout direciton calculations
function ActionBar:GetFlyoutDirection()
    local direction = self.sets.flyoutDirection or 'auto'

    if direction ~= 'auto' then
        return direction
    end

    local w, h = self:GetSize()
    local isVertical = w < h
    local anchor = self:GetPoint()

    if isVertical then
        if anchor and anchor:match('LEFT') then
            return 'RIGHT'
        end

        return 'LEFT'
    end

    if anchor and anchor:match('TOP') then
        return 'DOWN'
    end

    return 'UP'
end

function ActionBar:SetFlyoutDirection(direction)
    local oldDirection = self.sets.flyoutDirection or 'auto'
    local newDirection = direction or 'auto'

    if oldDirection ~= newDirection then
        self.sets.flyoutDirection = newDirection
        self:UpdateFlyoutDirection()
    end
end

function ActionBar:UpdateFlyoutDirection()
    self:ForButtons('SetFlyoutDirection', self:GetFlyoutDirection())
end

function ActionBar:Layout(...)
    ActionBar.proto.Layout(self, ...)

    self:UpdateFlyoutDirection()
end

function ActionBar:SaveFramePosition(...)
    ActionBar.proto.SaveFramePosition(self, ...)

    self:UpdateFlyoutDirection()
end

-- utility methods
function ActionBar:ForAll(method, ...)
    for _, bar in pairs(active) do
        bar:MaybeCallMethod(method, ...)
    end
end

-- the right click menu for an action bar
do
    local dropdownItems = {}
    local lastNumBars = -1

    local function getDropdownItems()
        local numBars = Addon:NumBars()

        if lastNumBars ~= numBars then
            dropdownItems = {
                {value = -1, text = _G.DISABLE}
            }

            for i = 1, numBars do
                dropdownItems[i + 1] = {
                    value = i,
                    text = ('Action Bar %d'):format(i)
                }
            end

            lastNumBars = numBars
        end

        return dropdownItems
    end

    local function AddStateGroup(panel, categoryName, stateType, l)
        local states =
            Addon.BarStates:map(
            function(s)
                return s.type == stateType
            end
        )

        if #states == 0 then
            return
        end

        panel:NewHeader(categoryName)

        for _, state in ipairs(states) do
            local id = state.id
            local name = state.text
            if type(name) == 'function' then
                name = name()
            elseif not name then
                name = l['State_' .. id:upper()]
            end

            panel:NewDropdown {
                name = name,
                items = getDropdownItems,
                get = function()
                    local offset = panel.owner:GetOffset(state.id) or -1
                    if offset > -1 then
                        return (panel.owner.id + offset - 1) % Addon:NumBars() + 1
                    end
                    return offset
                end,
                set = function(_, value)
                    local offset

                    if value == -1 then
                        offset = nil
                    elseif value < panel.owner.id then
                        offset = (Addon:NumBars() - panel.owner.id) + value
                    else
                        offset = value - panel.owner.id
                    end

                    panel.owner:SetOffset(state.id, offset)
                end
            }
        end
    end

    function ActionBar:CreateMenu()
        local menu = Addon:NewMenu(self.id)

        self:AddLayoutPanel(menu)
        self:AddPagingPanel(menu)
        menu:AddAdvancedPanel()
        menu:AddFadingPanel()

        self.menu = menu
    end

    function ActionBar:AddLayoutPanel(menu)
        local l = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
        local panel = menu:NewPanel(l.Layout)

        panel.sizeSlizer =
            panel:NewSlider {
            name = l.Size,
            min = 1,
            max = function()
                return panel.owner:MaxLength()
            end,
            get = function()
                return panel.owner:NumButtons()
            end,
            set = function(_, value)
                panel.owner:SetNumButtons(value)
                panel.colsSlider:UpdateValue()
            end
        }

        panel:AddLayoutOptions()

        return panel
    end

    function ActionBar:AddPagingPanel(menu)
        local l = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
        local panel = menu:NewPanel('Paging')

        AddStateGroup(panel, UnitClass('player'), 'class', l)
        AddStateGroup(panel, l.QuickPaging, 'page', l)
        AddStateGroup(panel, l.Modifiers, 'modifier', l)
        AddStateGroup(panel, l.Targeting, 'target', l)

        return panel
    end
end

local ActionBarsModule = Addon:NewModule('ActionBars', 'AceEvent-3.0')

function ActionBarsModule:Load()
    self:RegisterEvent('UPDATE_SHAPESHIFT_FORMS')
    self:RegisterEvent('UPDATE_BONUS_ACTIONBAR', 'UpdateOverrideBar')

    if _G.OverrideActionBar then
        self:RegisterEvent('UPDATE_VEHICLE_ACTIONBAR', 'UpdateOverrideBar')
        self:RegisterEvent('UPDATE_OVERRIDE_ACTIONBAR', 'UpdateOverrideBar')
    end

    self:RegisterEvent('PET_BAR_HIDEGRID')

    for i = 1, Addon:NumBars() do
        ActionBar:New(i)
    end
end

function ActionBarsModule:Unload()
    self:UnregisterAllEvents()

    ActionBar:ForAll('Free')
end

function ActionBarsModule:UpdateOverrideBar()
    if InCombatLockdown() or (not Addon.OverrideController:OverrideBarActive()) then
        return
    end

    Addon:GetOverrideBar():ForButtons('Update')
end

-- workaround for empty buttons not hiding when dropping a pet action
function ActionBarsModule:PET_BAR_HIDEGRID()
    if InCombatLockdown() then
        return
    end

    ActionBar:ForAll('HideGrid', ACTION_BUTTON_SHOW_GRID_REASON_EVENT)
end

function ActionBarsModule:UPDATE_SHAPESHIFT_FORMS()
    if InCombatLockdown() then
        return
    end

    ActionBar:ForAll('UpdateStateDriver')
end
